/*************************************************************************
 * AIZO CONFIDENTIAL
 * __________________
 *
 *  [2011] - [2013] AIZO AG
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Aizo AG. The intellectual and technical concepts
 * contained herein are proprietary to Aizo AG and may be covered
 * by European and Foreign Patents, patents in process, and are
 * protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from AIZO AG.
 * Authors: Sebastian Scholz <digitalstrom@sunshine-studios.org>
 *          Andreas Brauchli <andreas.brauchli@aizo.com>
 */

Ext.define('DSS.addon.Metering', {

    extend: 'DSS.addon.Framework',
    mixins: {
        observable: 'Ext.util.Observable'
    },

    uses: ['DSS.addon.Metering.MainPanel'],

    config: {
        appName: 'Metering',
        appId: 'system-addon-metering',
        appVersion: '2.0.1',
        appIcon: 'default_icon.png'
    },

    constructor: function() {
        this.callParent(arguments);
        this.addEvents('lastcalledscenesload');
        this.addEvents('chartrangechange');
    },

    getHelp: function() {
        return Ext.create('Ext.Component', {
            loader: {
                url: 'locale/' + dss.staticDataModel.activeLanguage + '/help.html',
                autoLoad: true,
                disableCaching: false
            }
        });
    },
    getContent: function() {
        return Ext.create("DSS.addon.Metering.MainPanel", Ext.apply({ id: 'metering-main-panel' }, this.config));
    }
});

Ext.onReady(function() {
    dss.buildUpLang(['locale/{languageSuffix}/dss-addon.po']);
    initContants();
    var metering = Ext.create('DSS.addon.Metering');
    metering.initPage();
});
/*######################################
 static configs
 ######################################*/
function initContants() {
    DSS.addon.Metering.time = {
        1: _("1 second"),
        2: _("1 second"),
        60: _("1 minute"),
        900: _("15 minutes"),
        86400: _("1 day"),
        604800: _("1 week"),
        2592000: _("1 month")
    };

    DSS.addon.Metering.chartMajorTicks = {
        1: 25,   // sec
        2: 25,   // 2sec
        60: 25,  // min
        900: 25, // 15min
        86400: 25,   // 24h
        604800: 25,  // 7 * 24h
        2592000: 25  // 30 * 24h
    };

    DSS.addon.Metering.chartSteps = {
        1: [Ext.Date.MINUTE, 1],
        2: [Ext.Date.MINUTE, 1],
        60: [Ext.Date.MINUTE, 15],
        900: [Ext.Date.HOUR, 1],
        86400: [Ext.Date.DAY, 7],
        604800: [Ext.Date.MONTH, 3],
        2592000: [Ext.Date.MONTH, 4]
    };

    DSS.addon.Metering.chartMinorTicks = {
        1: 5,   // sec
        2: 5,   // 2sec
        60: 2,  // min
        900: 3, // 15min
        86400: 6,   // 24h
        604800: 2,  // 7 * 24h
        2592000: 7  // 30 * 24h
    };

    DSS.addon.Metering.dateFormat = {
        1: 'G:i',
        2: 'G:i',
        60: 'G:i',
        900: 'G:i', // 900: 'j. M - G:i',
        86400: 'j. M Y',
        604800: 'j. M Y',
        2592000: 'j. M Y'
    };

    DSS.addon.Metering.minRefreshInterval = 2000;
    DSS.addon.Metering.YAxisDefaultMax = 20;
    DSS.addon.Metering.columnValueCount = 150;
}
/*######################################
 MainPanel
 ######################################*/
Ext.define('DSS.addon.Metering.MainPanel', {
    extend: 'Ext.panel.Panel',

    alias: 'widget.mainpanel',

    initComponent: function() {

        this.dSMPanel = Ext.create('DSS.addon.Metering.dSMPanel');
        this.centerPanel = Ext.create('DSS.addon.Metering.CenterPanel');

        Ext.apply(this, {
            split: true,
            border: true,
            layout: {
                type: 'border'
            },
            items: [this.dSMPanel, this.centerPanel]
        });

        this.callParent(arguments);
    }
});

/*######################################
 dSMPanel
 ######################################*/
Ext.define('DSS.addon.Metering.dSMPanel', {
    extend: 'Ext.panel.Panel',

    alias: 'widget.dsmpanel',

    initComponent: function() {

        this.dSMView = Ext.create('DSS.addon.Metering.dSMView', { flex: 9 });
        this.statsBox = Ext.create('Ext.Component', {
            html: '',
            cls: 'statsbox',
            flex: 0,
            isBlocked: false,
            clearStats: function() {
                this.getEl().setHTML('');
                this.isBlocked = false;
            },
            setStats: function(total, totalUnit, avg, avgUnit) {
                var t = _("Total: %total%unit").replace('%total', total).replace('%unit', totalUnit),
                    a = _("Average: %average%unit").replace('%average', avg).replace('%unit', avgUnit);
                this.isBlocked = true;
                this.getEl().setHTML(Ext.String.htmlEncode(t) + '<br>' + Ext.String.htmlEncode(a));
            },
            setItem: function(storeItem, unit) {
                if (this.isBlocked || !storeItem) {
                    return;
                }
                var d = _("Time: %date#d.m.Y H:i:s#"),
                    v = _("Value: %value%unit"),
                    dt = d.match(/%date#(.*)#/);

                d = d.replace(dt[0], Ext.Date.format(new Date(storeItem.get('date')), dt[1]));
                v = v.replace('%unit', unit)
                     .replace('%value', storeItem.get('value').toFixed(0));
                this.getEl().setHTML(Ext.String.htmlEncode(d) + '<br>' + Ext.String.htmlEncode(v));
            }
        });

        Ext.apply(this, {
            region: 'west',
            border: false,
            split: true,
            minWidth: 50,
            width: 200,
            layout: {
                type: 'vbox',
                align: 'stretch'
            },
            items: [this.dSMView, this.statsBox]
        });

        this.callParent(arguments);
    }
});

/*######################################
 dSMView
 ######################################*/
Ext.define('DSS.addon.Metering.dSMView', {
    extend: 'Ext.view.View',

    initComponent: function() {
        var me = this,
            store = Ext.create('Ext.data.Store', {
                model: 'DSS.addon.Metering.stores.CircuitModel'
            }),
            circuitTemplate;

        this.circuitStore = store;
        store.add({
            name: _("whole installation"),
            dsid: 0
        });
        store.on('load', function() {
            me.getSelectionModel().select(0);
        });
        store.suspendEvents(true); // queue events until after initComponent
        store.load({ addRecords: true });

        circuitTemplate = new Ext.XTemplate(
                '<tpl for=".">',
                    '<div class="zone-wrap" id="zone-{dsid}">',
                    '<span>', '{text}', '</span>',
                    '</div>',
                '</tpl>',
                '<div class="x-clear"></div>');

        Ext.apply(this, {
            title: _("dSM Overview"),
            store: this.circuitStore,
            tpl: circuitTemplate,
            autoScroll: true,
            overflow: 'auto',
            singleSelect: true,
            multiSelect: false,
            simpleSelect: false,
            trackOver: true,
            overItemClass: 'x-item-over',
            overClass: 'x-item-over',
            itemSelector: 'div.zone-wrap',
            emptyText: _("loading dSMs"),
            listeners: {
                viewready: function() {
                    var selModel = me.getSelectionModel();
                    if (selModel.getSelection().length === 0) {
                        selModel.select(0);
                    }
                },
                selectionchange: function(/*dv, nodes*/) {
                    var dsid = this.getSelectedDSID(),
                        cp = Ext.getCmp('metering-main-panel').centerPanel;
                    if (dsid !== -1) {
                        cp.setChartMeter(dsid);
                    }
                }
            }

        });
        this.callParent(arguments);
        store.resumeEvents();
    },
    /* Get the dsid of the selected dsm
     * If the store isn't loaded yet it will select the first element
     * when it's ready.
     * @return {String|Number} dsid, 0 if the whole apartment is chosen or -1 if none selected
     */
    getSelectedDSID: function() {
        var node = this.getSelectionModel().getSelection();
        if (node.length > 0) {
            return node[0].get('dsid');
        }
        return -1;
    },
    getSelectedName: function() {
        var node = this.getSelectionModel().getSelection();
        if (node.length > 0) {
            return node[0].get('text');
        }
        return -1;
    }
});

/*######################################
 CenterPanel
 ######################################*/

Ext.define('DSS.addon.Metering.CenterPanel', {
    extend: 'Ext.panel.Panel',

    listeners: {
        render: function(s) {
            var mouseEvt = function(evt) {
                    var delta = Math.max(-1, Math.min(1, (evt.wheelDelta || -evt.detail))),
                        sl = this.panSlider;

                    if (sl.isDisabled()) {
                        return;
                    }
                    sl.setValue(sl.getValue() + (Math.max(6, sl.increment) * delta));
                };
            // WebKit | IE9+
            Ext.get(s.id).dom.addEventListener('mousewheel', Ext.bind(mouseEvt, s), false);
            // FF
            Ext.get(s.id).dom.addEventListener('DOMMouseScroll', Ext.bind(mouseEvt, s), false);
        }
    },

    initComponent: function() {
        var me = this,
            zoomFieldContainer,
            incZoomButton,
            decZoomButton,
            typeBox,
            typeStore,
            chartTooltipRenderer = function(item) {
                me.statsBox.setItem(item.storeItem, me.unit);
            },
            chartBoxReady = function(chart) {
                var mask = chart.mask,
                    box = mask.getBox();
                // chart.mask is only constructed on afterrender event, boxready is after that and only thrown once
                // Careful: lots of private data access being done here (chart.mask / chart.maskSprite)
                mask.dd.destroy();
                mask.dd = { destroy: Ext.emptyFn }; // mask.dd.destroy is called on mask.destroy();
                mask.getEl().setStyle('cursor', 'default');
                mask.getEl().on('click', Ext.bind(me.hideChartSelectionMask, me));
                // chart.maskSprite.cursor = 'default'; // doesn't work so destroy the whole sprite and rebuild it
                chart.maskSprite.destroy();
                chart.maskSprite = chart.surface.add({ // values taken from Ext.Chart.MaskLayer source
                    type: 'path',
                    path: ['M', 0, 0],
                    zIndex: 1001,
                    opacity: 0.6,
                    hidden: true,
                    stroke: '#00f',
                    cursor: 'default' // changed from crosshair
                });
                // since we can't hook into the mask line position (or it's draw events as hacky fallback)
                // and the chart doesn't support vertical tracking of the data points, we'll do it ourself
                chart.getEl().on({
                    mousemove: function(e) {
                        var item = me.getStoreItemForPosition(e.getX() - chart.chartBBox.x);
                        me.statsBox.setItem(item, me.unit);
                    }
                });
            };

        //
        // Instance variable definitions
        //
        this.isLoading = true;
        this.initDone = false;
        this.resolutionStoreLoaded = false;
        this.jumpMarkers = [];
        this.lastCalledScenes = [];
        this.lastCalledScenesLoading = false;
        this.lastCalledScenesRequests = [];
        this.getLatestReq = null;

        this.unit = 'W';
        this.meterDsid = -1;
        this.chartType = ''; // can't set it as it won't load on first setChartType() otherwise
        this.resolution = 60; // set in setChartResolution
        this.timerHandle = null;
        this.refreshUrl = '';
        this.isShifted = false;
        this.zoomLevel = 0;
        this.stepRange = 0; // Number of total chartMajorTicks[resolution]
        this.statsBox = null; // Ext.getCmp('metering-main-panel').dSMPanel.statsBox;
        this.on('boxready', function() {
            me.statsBox = Ext.getCmp('metering-main-panel').dSMPanel.statsBox;
        });
        this.minJumpDev = 7; // Do not mark jump unless value deviation > 7
        // Factor-adjusted std. deviation. - currently 2.0 * stddev
        // (~95% of regular pts are withing this interval of the global average)
        this.stddev2 = 0;
        this.stddevAvg = 0; // Global average
        this.stddevSamples = 0; // Number of samples in the std.dev
        this.stddev2Local = 0;
        this.stddev2IgnoreLocalDev = 0;
        this.stddev2LocalSamples = 30;
        this.stddevLocalAvg = 0; // Local average
        this.stddevLocalAvgSamples = 0; // Number of samples in local average
        this.lastNewJump = null;

        this.resolutionStore = Ext.create('DSS.addon.Metering.stores.ResolutionStore');
        this.resolutionStore.on('load', function() { me.initStores(); });
        this.resolutionStoreFilter = {
            filterFn: function(item) {
                var type = item.get('type');
                return (!type || (!me.chartType && type === 'consumption') || me.chartType === type);
            }
        };

        typeStore = Ext.create('Ext.data.Store', {
            fields: ['type', 'name'],
            data: [{
                type: 'consumption',
                name: _("instantaneous consumption") + " (W)"
            }, {
                type: 'energyDelta',
                name: _("consumption") + " (Wh)"
            }]
        });

        typeBox = Ext.create('Ext.form.field.ComboBox', {
            store: typeStore,
            labelAlign: 'top',
            editable: false,
            displayField: 'name',
            valueField: 'type',
            fieldLabel: _("unit"),
            width: 200
        });
        typeBox.on('change', function(box, newVal) {
            me.setChartType(newVal);
        });
        typeBox.setValue('consumption');

        this.timeBox = Ext.create('Ext.form.field.ComboBox', {
            store: this.resolutionStore,
            labelAlign: 'top',
            editable: false,
            queryMode: 'remote',
            lastQuery: '',
            forceSelection: true,
            displayField: 'name',
            valueField: 'resolution',
            fieldLabel: _("resolution"),
            width: 100
        });
        this.timeBox.on('change', function(box, newVal) {
            me.setChartResolution(newVal);
        });

        this.zoomSlider = Ext.create('Ext.slider.Slider', {
            width: 100,
            value: 0,
            increment: 1,
            minValue: 0,
            maxValue: 10,
            useTips: false,

            reset: function() {
                me.zoomSlider.suspendEvents();
                me.zoomSlider.setValue(0);
                me.zoomSlider.resumeEvents();
            }
        });

        this.zoomSlider.on('change', function(slider, newVal) {
            var store = me.getMeteringStore(),
                chart = me.getMeteringChart(),
                // hacky but quick way to find the currently displayed items
                serieItems = chart.series.getAt(0).items,
                axisX = chart.axes.get(1),
                axisY = chart.axes.get(0);

            me.zoomLevel = newVal;
            if (store.getCount() === 0) {
                return;
            }
            // pick the second-last element as upper bound to avoid shifting to the next major tick
            me.setTimeAxisRange(axisX, store.first().get('date'), serieItems[serieItems.length-2].value[0]);
            me.setCurrentNumericAxisMax(axisX, axisY, store);
            me.adjustPanSlider();
            store.fireEvent('refresh', store); // HACK: same hack as panning to redraw whole line
            // me.getMeteringChart().refresh();
        });

        incZoomButton = Ext.create('Ext.button.Button', {
            cls: 'x-button-transparent',
            icon: 'framework/images/buttonIcons/add.gif',
            handler: function() {
                me.zoomSlider.setValue(me.zoomSlider.getValue() + 4);
            }
        });

        decZoomButton = Ext.create('Ext.button.Button', {
            cls: 'x-button-transparent',
            icon: 'framework/images/dss/delete.png',
            handler: function() {
                me.zoomSlider.setValue(me.zoomSlider.getValue() - 4);
            }
        });

        zoomFieldContainer = Ext.create('Ext.form.FieldContainer', {
            fieldLabel: _("Zoom"),
            labelAlign: 'top',
            layout: {
                type: 'hbox',
                align: 'stretch'
            },
            items: [decZoomButton, me.zoomSlider, incZoomButton]
        });

        this.panSlider = Ext.create('Ext.slider.Slider', {
            width: '100%',
            value: 100,
            // increment: 3,
            minValue: 10, // don't go back all the way since this only causes zooming
            maxValue: 100,
            cls: 'x-slider-background',
            useTips: false,
            listeners: {
                change: Ext.bind(me.onPanSliderChange, me)
            }
        });

        this.meteringLineStore = Ext.create('DSS.addon.Metering.MeteringStore');
        this.meteringLineStore.on('beforeload', function() {
            me.lineChart.setLoading(true);
        });
        this.meteringLineStore.on('load', Ext.bind(this.onMeteringStoreLoad, this));
        // Load will be called by setChartType as soon as default selections are made

        this.meteringColumnStore = Ext.create('DSS.addon.Metering.MeteringStore');
        this.meteringColumnStore.on('beforeload', function() {
            me.columnChart.setLoading(true);
        });
        this.meteringColumnStore.on('load', Ext.bind(this.onMeteringStoreLoad, this));

        this.lineChart = Ext.create('Ext.chart.Chart', {
            store: this.meteringLineStore,
            // autoScroll: true, // doesn't go well with chrome which add scrollbars on selection drag
            enableMask: true, // hidden option
            mask: 'horizontal',
            axes: [
                {
                    type: 'Numeric',
                    position: 'left',
                    fields: ['value'],
                    decimals: 0,
                    maximum: DSS.addon.Metering.YAxisDefaultMax,
                    minimum: 0,
                    grid: true,
                    title: _("instantaneous consumption") + " (W)"
                }, {
                    type: 'Time',
                    constrain: true, // do not draw before axis-start
                    position: 'bottom',
                    fields: ['date'],
                    // dateFormat: DSS.addon.Metering.dateFormat[60],
                    label: { renderer: Ext.bind(me.dateFormatLabelRenderer, me) },
                    dashSize: 6,
                    grid: true,
                    minorTickSteps: DSS.addon.Metering.chartMinorTicks[60],
                    step: DSS.addon.Metering.chartSteps[60],
                    adjustMinimumByMajorUnit: true,
                    adjustMaximumByMajorUnit: true
                }
            ],
            series: [{
                type: 'Line',
                xField: 'date',
                yField: 'value',
                axis: ['left', 'bottom'],
                smooth: false,
                showMarkers: false,
                highlight: false,
                style: {
                    fill: '#00523b',
                    stroke: '#00523b',
                    'stroke-width': 3
                },
                selectionTolerance: 5,
                listeners: {
                    itemmouseover: chartTooltipRenderer
                }
            }],
            theme: 'MeteringTheme'
        });
        this.lineChart.on('refresh', function() {
            if (!me.meteringLineStore.isLoading()) {
                me.updateMarkers();
                return true;
            }
            return false;
        });
        this.lineChart.on('resize', function() {
            if (!me.meteringLineStore.isLoading()) {
                me.updateMarkers();
                return true;
            }
            return false;
        });
        this.lineChart.on('boxready', chartBoxReady);
        this.on('chartrangechange', Ext.bind(me.hideChartSelectionMask, me));
        this.lineChart.on('click', Ext.bind(me.hideChartSelectionMask, me));
        this.lineChart.on('select', Ext.bind(me.onChartSelection, me));

        this.columnChart = Ext.create('Ext.chart.Chart', {
            store: this.meteringColumnStore,
            // autoScroll: true, // doesn't go well with chrome which add scrollbars on selection drag
            axes: [{
                type: 'Numeric',
                position: 'left',
                fields: ['value'],
                decimals: 0,
                maximum: DSS.addon.Metering.YAxisDefaultMax,
                minimum: 0,
                grid: true,
                title: _("consumption") + " (Wh)"
            }, {
                type: 'Time',
                constrain: true, // do not draw before axis-start
                position: 'bottom',
                fields: ['date'],
                // dateFormat: this.getDateFormat(), //DSS.addon.Metering.dateFormat[60],
                label: { renderer: Ext.bind(me.dateFormatLabelRenderer, me) },
                dashSize: 6,
                grid: true,
                minorTickSteps: DSS.addon.Metering.chartMinorTicks[60],
                step: DSS.addon.Metering.chartSteps[60],
                adjustMinimumByMajorUnit: true,
                adjustMaximumByMajorUnit: true
            }],
            series: [{
                type: 'Column',
                xField: 'date',
                yField: 'value',
                axis: ['left', 'bottom'],
                highlight: false,
                style: {
                    fill: '#00523b',
                    stroke: '#00523b',
                    'stroke-width': 3
                },
                listeners: {
                    itemmouseover: chartTooltipRenderer
                }/*,
                // Fixed width columns
                renderer: function(sprite, record, attr, index, store) {
                    return Ext.apply(attr, {width:10});
                }*/
            }],
            enableMask: true, // hidden option
            mask: 'horizontal',
            hidden: true,
            theme: 'MeteringTheme'
        });
        this.columnChart.on('beforerefresh', function() {
            return (!me.meteringColumnStore.isLoading());
        });
        this.columnChart.on('resize', function() {
            return (!me.meteringColumnStore.isLoading());
        });
        this.columnChart.on('boxready', chartBoxReady);
        this.columnChart.on('click', Ext.bind(me.hideChartSelectionMask, me));
        this.columnChart.on('select', Ext.bind(me.onChartSelection, me));

        Ext.apply(this, {
            layout: 'fit',
            items: [this.lineChart, this.columnChart],
            tbar: [
                typeBox,
                this.timeBox,
                { xtype: 'tbseparator' },
                zoomFieldContainer,
                '->',
                {
                    xtype: 'button',
                    text: _("clear screen"),
                    iconCls: 'icon-clear',
                    handler: function() {
                        me.clearMarkers();
                        me.getMeteringStore().removeAll();
                        me.stddev2IgnoreLocalDev = me.stddev2LocalSamples;
                    }
                }, {
                    xtype: 'button',
                    text: _("export as CSV"),
                    iconCls: 'icon-export',
                    handler: function() {
                        var dSMName = Ext.getCmp('metering-main-panel').dSMPanel.dSMView.getSelectedName(),
                            date = Ext.Date.format(new Date(), 'Y-m-d'),
                            // a temporary form has to be used so that the download
                            tempForm = document.createElement('form'),
                            params = {
                                Filename: dSMName + '_' + date + '.csv',
                                Contenttype: 'text/csv',
                                Data: me.createCSV()
                            },
                            par,
                            opt;

                        tempForm.action = '/cgi-bin/pushfile.hs';
                        tempForm.method = 'POST';
                        tempForm.style.display = 'none';
                        for(par in params) {
                            opt = document.createElement('input');
                            opt.type = 'hidden';
                            opt.name = par;
                            opt.value = params[par];
                            tempForm.appendChild(opt);
                        }
                        document.body.appendChild(tempForm);
                        tempForm.submit();
                        document.body.removeChild(tempForm);
                    }
                }
            ],
            bbar: [this.panSlider],
            region: 'center',
            border: false,
            split: true,
            minWidth: 600
        });

        // call manually since resolutionStore is now cached the load event
        // won't be triggered.
        this.initStores();
        this.timeBox.setValue(this.resolutionStore.getAt(1));

        me.callParent(arguments);
        this.initDone = true;
        this.updateLoadingStatus();
    },
    getMeteringChart: function() {
        if (this.chartType === 'consumption') {
            return this.lineChart;
        }
        return this.columnChart;
    },
    getMeteringStore: function() {
        if (this.chartType === 'consumption') {
            return this.meteringLineStore;
        }
        return this.meteringColumnStore;
    },
    initStores: function() {
        dss.staticDataModel.initNameCache();
        dss.staticDataModel.initDeviceCache();
        this.resolutionStore.filter(this.resolutionStoreFilter);
        this.resolutionStoreLoaded = true;
        this.updateLoadingStatus();
    },
    updateLoadingStatus: function() {
        this.isLoading = !(this.resolutionStoreLoaded && this.initDone);
        if (!this.isLoading) {
            this.setChartType('consumption'); // load the metering store
        }
    },
    /**
     * Called directly from handleGetLatest
     * @param data {Array} Array of one element with one record
     *             formed: [ date, value ]
     */
    onLiveMeteringDataAdded: function(storeItem) {
        var store = this.getMeteringStore(),
            chartAxes = this.getMeteringChart().axes,
            axisX = chartAxes.get(1),
            axisY = chartAxes.get(0),
            last = storeItem.get('time'),
            value = storeItem.get('value');

        if (!this.isShifted) {
            // Adjust Y-axis
            if (value * 1.2 > axisY.maximum) {
                this.setNumericAxisMax(axisY, value);
            }
            // Adjust X-axis
            if (this.setTimeAxisRange(axisX, store.first().get('date'), last)) {
                this.adjustPanSlider();
                this.fireEvent('chartrangechange');
            }
        }
    },
    dateFormatLabelRenderer: function(v) {
        return Ext.Date.format(new Date(v), DSS.addon.Metering.dateFormat[this.resolution]);
    },
    setChartMeter: function(dsid) {
        if (this.isLoading) {
            this.meterDsid = dsid;
            return;
        }
        if (this.meterDsid !== dsid) {
            this.meterDsid = dsid;
            this.stopTimer();
            // reload the store
            this.loadMeteringStore();
        }
    },
    /**
     * Set the chart resolution
     * @param resolution {Number} Resolution to set in seconds
     */
    setChartResolution: function(resolution) {
        if (this.isLoading) {
            return;
        }

        if (this.resolution !== resolution) {
            // chart resolution changed
            this.resolution = resolution;
            this.stopTimer();
            this.zoomSlider.reset();
            this.internalSetChartResolution(resolution);
            this.loadMeteringStore();
        }
    },
    /** @private */
    internalSetChartResolution: function(resolution) {
        var axisX = this.getMeteringChart().axes.get(1);
        axisX.step = DSS.addon.Metering.chartSteps[resolution];
        axisX.minorTickSteps = DSS.addon.Metering.chartMinorTicks[resolution];
    },
    setChartType: function(type) {
        if (this.isLoading) {
            return;
        }
        var isConsumption = (type === 'consumption'),
            oldStore = this.getMeteringStore(),
            resolution;

        if (this.chartType !== type) {
            this.stopTimer();
            oldStore.suspendEvents();
            oldStore.removeAll(); // clear current store - delay events
            //chart type has to be changed
            this.chartType = type;
            this.zoomSlider.reset();
            this.unit = isConsumption ? 'W' : 'Wh';
            this.resolutionStore.clearFilter(true);
            this.resolutionStore.filter(this.resolutionStoreFilter);
            //only allow selected resolution values that are not filtered in the store
            if (this.resolutionStore.findExact('resolution', this.resolution) === -1) {
                resolution = this.resolutionStore.getAt(0);
                this.timeBox.suspendEvents();
                this.timeBox.setValue(resolution);
                this.timeBox.resumeEvents();
                this.resolution = resolution.get('resolution');
                this.internalSetChartResolution(this.resolution);
            }
            this.loadMeteringStore();

            this.lineChart.setVisible(isConsumption);
            this.columnChart.setVisible(!isConsumption);
            oldStore.resumeEvents();
        }
    },
    /**
     * Load metering store
     */
    loadMeteringStore: function() {
        if (this.meterDsid === -1) {
            return;
        }

        var store = this.getMeteringStore(),
            url = '/json/metering/getValues?';

        if (this.getLatestReq) {
            Ext.Ajax.abort(this.getLatestReq);
        }
        this.abortLastCalledScenesCall();
        this.clearMarkers();

        url += (this.meterDsid === 0 ? 'dsid=.meters(all)' : 'dsid='+ this.meterDsid);
        url += '&type=' + this.chartType
             + '&resolution=' + this.resolution
             + '&valueCount=600';
        store.getProxy().url = url;
        store.load();

        if (this.resolution < 10) {
            // Refresh lastCalledScenes
            this.requestLastCalledScenes();
        }
    },
    createCSV: function() {
        var store = this.getMeteringStore(),
            count = store.getCount(),
            csvString = '',
            i, r;

        csvString = '"' + _("Time") + '";' + '"';
        csvString += (this.chartType === 'consumption' ?  _("instantaneous consumption") : _("consumption"));
        csvString += ' (' + _("in %unit").replace('%unit', this.unit) + ')"' + '\n';
        for (i = 0; i < count; ++i) {
            r = store.getAt(i);
            csvString += '"' + Ext.Date.format(r.get('date'), 'd.m.Y G:i:s')
                        + '";' + r.get('value') + '\n';
        }
        return csvString;
    },
    /**
     * Start periodic timer, fetching live data
     */
    startTimer: function() {
        if (this.resolution > 2 * 60 * 60) {
            // don't schedule anything above 2h to avoid 32bit timer overflows
            return;
        }

        if (this.resolution <= 2 && this.chartType === 'consumption') {
            this.refreshUrl = '/json/metering/getAggregatedLatest?type=consumption'
                + '&from=.meters('
                + (this.meterDsid === 0 ? 'all' : this.meterDsid)
                + ')';
        } else {
            this.refreshUrl = '/json/metering/getAggregatedValues?from='
                + (this.meterDsid === 0 ? 'all' : this.meterDsid)
                + '&type=' + this.chartType
                + '&resolution=' + this.resolution
                + '&valueCount=1';
        }

        this.lastTick = (new Date()).getTime();
        this.rescheduleTimer();
    },
    /**
     * Periodic timer function executed at every tick
     */
    timerTick: function() {
        // jumps in the data series are detected and markers added
        this.getLatestReq = Ext.Ajax.request({
            url: this.refreshUrl,
            success: Ext.bind(this.handleGetLatest, this)
        });
        this.lastTick = (new Date()).getTime();
    },
    rescheduleTimer: function() {
        var timer = Math.max(1000 * this.resolution, DSS.addon.Metering.minRefreshInterval);
        // deduce processing time for last request
        timer = Math.max(0, timer - ((new Date()).getTime() - this.lastTick));
        this.timerHandle = setTimeout(Ext.bind(this.timerTick, this), timer);
    },
    /**
     * Stop the periodic realtime data fetcher task
     */
    stopTimer: function() {
        if (this.timerHandle !== null) {
            clearTimeout(this.timerHandle); // complement to native setTimeout function
            this.timerHandle = null;
        }
    },
    handleLastCalledScenes: function(response, request, marker) {
        var jsonData;

        this.lastCalledScenesLoading = false;
        this.lastCalledScenesRequests = [];
        try {
            jsonData = Ext.JSON.decode(response.responseText);
            if (!jsonData.ok) { return; }
            jsonData = jsonData.result.value;

            this.lastCalledScenes = Ext.JSON.decode(jsonData);
            this.fireEvent('lastcalledscenesload', this.lastCalledScenes);
            if (marker) {
                this.matchLastCalledScene(marker);
            }
        } catch (e) { }
    },
    /**
     * Match a timestamp to a last called event of a given meter
     * @param meterDsid {String} meter or number 0 if all
     */
    findLastCalledSceneEvent: function(time, meterDsid) {
        var timeRange = this.resolution < 10 ? 10000 : 90000,
            rangeStart = new Date(time),
            rangeEnd = new Date(time),
            bestMatch,
            matchedEvent,
            sceneEvent = {},
            eventTime,
            setIf = function(evt, prop, i) {
                if (isNaN(i)) { return false; }
                evt[prop] = i;
                return true;
            },
            ls, tmp, i;

        // Check if last scene call was made within +/- timeRange sec of the storeItem
        rangeStart.setTime(rangeStart.getTime() - timeRange);
        rangeEnd.setTime(rangeEnd.getTime() + timeRange);
        bestMatch = [-1, Infinity];
        // find one matching entry from lastCalledScene, starting at the end
        for (i = this.lastCalledScenes.length - 1; i >= 0; --i) {
            ls = this.lastCalledScenes[i];
            if (meterDsid !== 0 && ls.meterDSID !== undefined && meterDsid !== ls.meterDSID) {
                continue;
            }

            eventTime = new Date(ls.time);
            if (eventTime < rangeStart) {
                break; // assuming ordered lastCalledScenes
            }
            if (eventTime <= rangeEnd) {
                tmp = eventTime.getTime() - time.getTime();
                if (tmp < 0) {
                    // slightly bias (0.5s) towards events that originated before the jump
                    // as the events arrive at the dSS and only then are dispatched
                    // EXCEPT for local-buttons ("Lokaltaster"), in which case there
                    // is at least 240ms delay for the event to reach the dSS after the
                    // device activated - this case is not handled here
                    tmp += Math.min(-tmp, 500);
                }
                tmp = tmp * tmp;
                if (tmp < bestMatch[1]) {
                    if (this.resolution >= 10 && bestMatch[0] !== -1) {
                        // resolution is too coarse to allow good matching with
                        // multiple possible matches -> discard
                        return false;
                    }
                    bestMatch = [i, tmp];
                }
            }
        }
        if (bestMatch[0] === -1) {
            return false;
        }
        // sceneEvent = this.lastCalledScenes.splice(bestMatch[0], 1); // Remove the match
        matchedEvent = this.lastCalledScenes[bestMatch[0]];

        setIf(sceneEvent, 'zoneId', parseInt(matchedEvent.zoneID, 10));
        setIf(sceneEvent, 'groupId', parseInt(matchedEvent.groupID, 10));
        if (matchedEvent.originDSID) {
            sceneEvent.originDsid = matchedEvent.originDSID;
            sceneEvent.originName = dss.staticDataModel.getDevice(sceneEvent.originDsid).name;
            sceneEvent.originIcon = dss.staticDataModel.getImageForDevice({dSID: sceneEvent.originDsid});
        }
        if (setIf(sceneEvent, 'sceneId', parseInt(matchedEvent.sceneID, 10))) {
            sceneEvent.sceneName = dss.staticDataModel.getNameForScene(sceneEvent.zoneId, sceneEvent.groupId, sceneEvent.sceneId);
            sceneEvent.sceneIcon = dss.staticDataModel.getImagePathForScene(sceneEvent.zoneId, sceneEvent.groupId, sceneEvent.sceneId);
        }
        if (matchedEvent.eventName) {
            sceneEvent.eventName = matchedEvent.eventName;
        }

        if (sceneEvent.sceneName || sceneEvent.eventName) {
            sceneEvent.zoneName = dss.staticDataModel.getNameForZone(sceneEvent.zoneId);
        }
        return sceneEvent;
    },
    /**
     * Match a marker/markers with cached lastCalledScenes
     * @param markers {Object/Array} Either a DiffSprite marker or an array thereof
     */
    matchLastCalledScene: function(markers) {
        var sceneEvent,
            i;

        if (!Ext.isArray(markers)) {
            markers = [markers];
        }
        for (i = 0; i < markers.length; ++i) {
            sceneEvent = this.findLastCalledSceneEvent(markers[i].storeItem.get('date'), this.meterDsid);
            if (sceneEvent) {
                markers[i].setSceneEvent(sceneEvent);
            } else {
                markers[i].fireEvent('sceneeventchange', markers[i], false);
            }
        }
    },
    /**
     * Called when the store has been (re)loaded
     * This function also refreshes the chart and starts the refresh timer
     */
    onMeteringStoreLoad: function() {
        var store = this.getMeteringStore(),
            chart = this.getMeteringChart(),
            count = store.getCount(),
            timeAxis = chart.axes.get(1),
            fromDate,
            toDate;

        // Adjust axes
        if (count > 0) {
            fromDate = store.first().get('date');
            toDate = store.last().get('date');
            this.setNumericAxisMax(chart.axes.get(0), store.max('value'));
            this.setTimeAxisRange(timeAxis, fromDate, toDate);
        } else {
            this.setTimeAxisRange(timeAxis);
        }
        this.resetPanSlider();
        // chart.refresh();
        // HACK: instead of chart refresh, trigger a refresh from the store to
        // get right values. I don't really see why, because the "wrong" data
        // on initial load (aggregated meters) seem to be the ones from a single
        // dSS (which is never loaded before!) - might be related to the pan
        // slider hack which is basically the same
        store.fireEvent('refresh', store);
        if (count > 0 && this.chartType === 'consumption' && this.resolution < 10) {
            // We can do jump detection on all resolutions, but event coupling doesn't
            // work too well for anything > 10s..
            this.detectJumps();
        }
        chart.setLoading(false);
        this.startTimer();
    },
    parseDate: function(dateString) {
        var dm = dateString.split(' '),
            dt = dm[0].split('-'),
            tm = dm[1].split(':');
        return new Date(dt[0], dt[1] -1, dt[2], tm[0], tm[1], tm[2]);
    },
    handleGetLatest: function(response) {
        var store = this.getMeteringStore(),
            jsonData,
            item,
            res,
            date,
            data;

        this.getLatestReq = null;
        try {
            jsonData = Ext.JSON.decode(response.responseText);
            if (!jsonData.ok) {
                return;
            }
            item = jsonData.result.values[0];
            if (!item) {
                return; // sometimes the answer is empty
            }
            if (this.resolution > 2) {
                res = parseInt(jsonData.result.resolution, 10);
                date = Ext.Date.parse(item[0], 'timestamp');
                data = [[date, item[1]]]; // Array of array
            } else {
                date = this.parseDate(item.date);
                data = [[date, item.value]]; // Array of array
            }
        } catch (e) {
            return;
        }

        // delay events until we updated everything
        // then the chart is only refreshed once
        store.suspendEvents(true);
        store.loadData(data, true /* append */);
        // loadData fires datachanged and refresh, not load!
        this.onLiveMeteringDataAdded(store.last());
        store.resumeEvents();
        if (this.chartType === 'consumption' && this.resolution < 60) {
            this.detectNewJump(store.last());
        }
        this.rescheduleTimer();
    },
    getStddev: function(records, avg) {
        var i = 0,
            len = records.length,
            sqsum = 0,
            tmp;

        for (i = 0; i < len; ++i) {
            tmp = avg - records[i].get('value');
            sqsum += tmp * tmp;
        }
        return len > 0 ? Math.sqrt(sqsum / len) : 0;
    },
    isDeviating: function(val, stdavg, stddev) {
        var dev = Ext.isNumber(stddev) ? stddev : this.stddev2,
            avg = Ext.isNumber(stdavg) ? stdavg : this.stddevAvg;
        return Math.abs(avg - val) > Math.max(dev, this.minJumpDev);
    },
    /**
     * detect jumps
     * the parameters of this function are storeItems
     * jump detection is based on all store values
     * (maybe refine with a max of 60 vals? (~10%))
     */
    detectJumps: function() {
        var me = this,
            findMaxima = function(items) {
                var store = me.getMeteringStore(),
                    len = items.length,
                    startIdx = len > 0 ? items[0].index : 0,
                    val = len > 0 ? store.getAt(startIdx).get('value') : 0,
                    prevVal,
                    maxima = [],
                    i;

                me.stddevAvg = store.getAverage(items, 'value');
                me.stddev2 = 2.0 * me.getStddev(items, me.stddevAvg);
                for (i = startIdx; i < len; ++i) {
                    prevVal = val;
                    val = store.getAt(i).get('value');
                    if (me.isDeviating(val)) {
                        maxima.push({idx: i, value: val, diff: val - prevVal});
                    }
                }
                return maxima;
            },
            segmentMaxima = function(maxima, startIdx) {
                var store = me.getMeteringStore(),
                    count = store.getCount(),
                    rangeStart,
                    prevIdx,
                    avg,
                    avgCount = 1,
                    segments = [],
                    pushMaximum = function(index) {
                        var tmp = store.getAt(index).get('value'),
                            diff = (index > 0 ? tmp - store.getAt(index -1).get('value') : 0),
                            rangeStart = { idx: index, value: tmp, diff: diff };
                            segments.push(rangeStart);
                    },
                    i;

                if (count === 0) {
                    return [];
                }
                startIdx = Ext.isNumber(startIdx) ? startIdx : 0;
                if (maxima.length === 0 || maxima[0].idx !== 0) {
                    // First segment from 0 to maxima[0].idx
                    pushMaximum(startIdx);
                    if (maxima.length === 0) {
                        return segments;
                    }
                }
                segments.push(maxima[0]);
                prevIdx = maxima[0].idx;
                avg = maxima[0].value; // use first maxima's value as average

                for (i = 1; i < maxima.length; ++i) {
                    if (maxima[i].idx === prevIdx +1 && !me.isDeviating(maxima[i].value, avg)) {
                        // same maximum, adapt average
                        ++prevIdx;
                        avg = (avg * avgCount + maxima[i].value) / (++avgCount);
                    } else {
                        if (maxima[i-1].idx +1 !== maxima[i].idx) {
                            // inject range between maxima, since we have a non-maximum ahead
                            // that is not already the next (different) maximum
                            pushMaximum(maxima[i-1].idx +1);
                        }

                        rangeStart = maxima[i];
                        avg = rangeStart.value; // use next maxima's value as average
                        avgCount = 1;
                        prevIdx = rangeStart.idx;
                        segments.push(rangeStart);
                    }
                }
                if (avgCount > 1 && maxima[i-1].idx +1 < count) {
                    // inject next non-maximal segment if not already done so by the 'else' case
                    pushMaximum(maxima[i-1].idx +1);
                }
                return segments;
            },
            refineSegments = function(segments) {
                var store = me.getMeteringStore(),
                    count = store.getCount(),
                    startIdx = segments[0].idx,
                    val, lastVal,
                    i, j, s;

                // start with 2nd segment since the first is always the fixed beginning of data
                for (i = (startIdx > 0 ? 0 : 1); i < segments.length; ++i) {
                    s = segments[i];
                    startIdx = s.idx;
                    // look ahead up to 5 pts
                    j = startIdx;
                    lastVal = s.value;
                    while (++j < count && j < startIdx + 5) {
                        val = store.getAt(j).get('value');
                        if ((val - (s.value + 0.25 * s.diff)) * (s.diff > 0 ? 1 : -1) > 0) {
                            s.idx += 1;
                            s.diff += val - lastVal; // shift point and add additional difference to jump
                            s.value = val;
                            lastVal = val;
                        } else {
                            break;
                        }
                    }
                    if (i > 0 && j > startIdx +1) {
                        // intermediate asc/desc values are not associated with a range
                        segments[i-1].endIdx = startIdx -1;
                    }
                    // look back if we can absorb one point
                    if (startIdx > 1) {
                        val = s.value - s.diff; // same as store.getAt(startIdx -1).get('value')
                        lastVal = store.getAt(startIdx -2).get('value');
                        if ((lastVal - (val + 0.25 * s.diff)) * (s.diff > 0 ? -1 : 1) > 0) {
                            s.diff += val - lastVal;
                            if (i > 0) {
                                segments[i-1].endIdx = startIdx -2;
                            }
                        }
                    }

                    // remove following overlapping segments or this and following segments with too small diff
                    j = i;
                    while (++j < segments.length && (s.idx >= segments[j].idx || Math.abs(segments[j].diff) < me.minJumpDev));
                    segments.splice(i+1, j-i -1);
                    if (Math.abs(s.diff) < me.minJumpDev) {
                        segments.splice(i, 1);
                        i -= 1;
                    }
                }

                if (segments.length > 0 && segments[0].idx > 0 && Math.abs(segments[0].diff) < me.minJumpDev) {
                    segments.shift();
                }
                return segments;
            },
            createJumpMarkers = function(segments) {
                var store = me.getMeteringStore(),
                    isErratic = function(segIdx) {
                        var s = segments[segIdx],
                            dev = Math.max(me.stddev2, me.minJumpDev),
                            len = (s.endIdx ? s.endIdx
                                   : segIdx < segments.length -1
                                       ? segments[segIdx+1].idx
                                       : store.getCount() -1) - s.idx;
                        // only erratic if < 3 pts and within 3 * stddev2 or not within stddev2 of avg
                        return len < 3 && (Math.abs(s.diff) < (3 * dev)
                                           || Math.abs(s.value - me.stddevAvg) <= dev)
                                || Math.abs(s.value - (s.diff < 0 ? s.diff : 0) - store.getAverage(store.getRange(s.idx, s.idx + len), 'value')) > dev;
                    },
                    confirmedSegments = [segments[0]],
                    storeItem,
                    sceneEvent,
                    i, s;

                // skip first segment (start of data)
                for (i = 1; i < segments.length; ++i) {
                    s = segments[i];
                    storeItem = store.getAt(s.idx);
                    if (Math.abs(s.diff) >= me.minJumpDev) {
                        sceneEvent = me.findLastCalledSceneEvent(storeItem.get('date'), me.meterDsid);
                        if (sceneEvent || !isErratic(i)) {
                            me.addJumpMarker(storeItem, s.diff, sceneEvent);
                            confirmedSegments.push(s);
                        }
                    }
                }
                return confirmedSegments;
            },
            createMarkersAndRerun = function(segments) {
                var subsegments = 0,
                    lastSegments = [],
                    lastSegIdx,
                    endIdx,
                    dataRange,
                    i, s;

                segments = createJumpMarkers(segments);
                for (i = 0; i < segments.length; ++i) {
                    // second pass within each segment
                    s = segments[i];
                    endIdx = i < segments.length -1 ? segments[i+1].idx -1 : store.getCount() -1;
                    dataRange = store.getRange(s.idx, endIdx);
                    lastSegments = refineSegments(segmentMaxima(findMaxima(dataRange), s.idx), s.idx);
                    subsegments += createJumpMarkers(lastSegments).length -1;
                }
                // update with last segment information for live detection
                lastSegIdx = lastSegments.length > 0 ? lastSegments[lastSegments.length -1].idx : -1;
                updateStdDev(lastSegIdx >= segments[segments.length -1].idx ? lastSegments : segments);
            },
            updateStdDev = function(segments) {
                if (segments.length === 0) {
                    return;
                }
                var store = me.getMeteringStore(),
                    count = store.getCount(),
                    s = segments[segments.length -1],
                    dataRange = store.getRange(s.idx),
                    localRange = count > me.stddev2LocalSamples ? store.getRange(count - me.stddev2LocalSamples - 1) : [],
                    localAvg = store.getAverage(localRange, 'value');

                me.stddev2Local = 2.0 * me.getStddev(localRange, localAvg);
                // update stddev with last segment's range
                me.stddevLocalAvg = store.last().get('value');
                me.stddevLocalAvgSamples = 1;
                me.stddevSamples = dataRange.length;
                // adapt global avg to 9/10 of local one
                me.stddevAvg = (9 * store.getAverage(dataRange, 'value') + store.average('value')) / 10;
                me.stddev2 = 2.0 * me.getStddev(dataRange, me.stddevAvg);
                me.stddev2 = (9.0 * me.stddev2 + me.stddev2Local) / 10.0;
            },
            // Careful, store.data.items is private
            store = this.getMeteringStore(),
            segments = refineSegments(segmentMaxima(findMaxima(store.data.items)));

        if (me.lastCalledScenesLoading) {
            me.on('lastcalledscenesload', Ext.bind(createMarkersAndRerun, me, [segments]), me, {single: true});
        } else {
            createMarkersAndRerun(segments);
        }
    },
    abortLastCalledScenesCall: function() {
        var markers = [];
        while (this.lastCalledScenesRequests.length > 0) {
            lr = this.lastCalledScenesRequests.pop();
            Ext.Ajax.abort(lr[0]);
            markers = markers.concat(lr[1]);
        }
        return markers;
    },
    /**
     * Request the lastCalledScene for given marker/markers
     * If requests are pending, these are aborted and attached to this call
     * @param marker {Object/Array} Either a DiffSprite marker or an array of markers
     */
    requestLastCalledScenes: function(marker) {
        var me = this,
            req,
            lr,
            markers = (marker ? (Ext.isArray(marker) ? marker : [marker]) : []);

        markers.concat(this.abortLastCalledScenesCall());
        this.lastCalledScenesLoading = true;
        req = Ext.Ajax.request({
            url: '/json/property/getString?path=/scripts/system-addon-metering/lastCalledScenes',
            success: Ext.bind(me.handleLastCalledScenes, me, markers, true)
        });
        me.lastCalledScenesRequests.push([req, markers]);
    },
    detectNewJump: function(storeItem) {
        var me = this,
            dev = this.stddev2 / 2.0,
            len = this.stddevSamples,
            store = this.getMeteringStore(),
            count = store.getCount(),
            val = storeItem.get('value'),
            diff = (count > 1) ? val - store.getAt(count -2).get('value') : 0,
            isDeviating = this.isDeviating(val, this.stddevLocalAvg)
                            && Math.abs(diff) >= Math.max(this.minJumpDev, this.stddev2),
            localRange = count > this.stddev2LocalSamples ? store.getRange(count - this.stddev2LocalSamples -1) : [],
            localAvg = store.getAverage(localRange, 'value'),
            point,
            jm,
            markerUpdated = false,
            sceneLoadedHandler = function(marker, sceneEventOrItem) {
                if (sceneEventOrItem !== false) {
                    // largely adapt the global average to the new local on (9/10)
                    this.stddevAvg = (9 * this.stddevLocalAvg + this.stddevAvg) / 10;
                    this.stddev2 = Math.min(this.stddev2, Math.abs(diff / 2));
                    marker.on('storeitemchange', Ext.bind(sceneLoadedHandler, this));
                }
            },
            continueLastJump = function(isDeviating) {
                var lastVal = (count > 1 ? store.getAt(count -2).get('value') : undefined),
                    lastDiff = (count > 2 ? lastVal - store.getAt(count -3).get('value') : 0),
                    isExtension = function(val, lastVal, lastDiff, fac) {
                        return (val - (lastVal + fac * lastDiff)) * (lastDiff >= 0 ? 1 : -1) > 0;
                    },
                    marker;

                if (me.lastNewJump && me.lastNewJump === store.getAt(count -2)) {
                    // storeItem.index === store.getCount() -1
                    // check if it's still a flank of the same jump
                    if (isExtension(val, lastVal, lastDiff, 0.10)) {
                        me.stddevLocalAvg = val;
                        marker = me.jumpMarkers[me.jumpMarkers.length -1];
                        diff = marker.getDiff() + diff;
                        // update
                        me.stddevLocalAvg = val;
                        marker.updateStoreItem(me.getPositionForStoreItem(storeItem), storeItem, diff);
                        return true;
                    }
                } else if (isDeviating && isExtension(val, lastVal, lastDiff, 0.4)) {
                    // look to expand backwards
                    diff += lastDiff;
                    jm = me.addJumpMarker(storeItem, diff);
                    if (me.resolution < 10) {
                        jm.on('sceneeventchange', Ext.bind(sceneLoadedHandler, me), { single: true });
                        setTimeout(me.requestLastCalledScenes.bind(me, [jm]), 1000);
                    }
                    return true;
                }
                return false;
            };

        if (isDeviating) {
            // new jump: reset average
            if (!continueLastJump(true)) {
                jm = this.addJumpMarker(storeItem, diff);
                if (this.resolution < 10) {
                    jm.on('sceneeventchange', Ext.bind(sceneLoadedHandler, this), { single: true });
                    setTimeout(this.requestLastCalledScenes.bind(this, [jm]), 1000);
                }
            }
            this.lastNewJump = storeItem;
            this.stddevLocalAvg = val;
            this.stddevLocalAvgSamples = 1;
            this.stddev2IgnoreLocalDev = this.stddev2LocalSamples;
        } else {
            if (continueLastJump()) {
                this.stddev2IgnoreLocalDev = this.stddev2LocalSamples;
            } else {
                // adjust local average
                this.stddevLocalAvg = (
                        (this.stddevLocalAvg * this.stddevLocalAvgSamples + val)
                        / (this.stddevLocalAvgSamples + 1));
                this.stddevLocalAvgSamples += 1;
            }
        }
        // adjust average and std deviation with new sample
        this.stddevAvg = (this.stddevAvg * len + val) / (len + 1);
        val -= this.stddevAvg;
        this.stddev2Local = 2.0 * this.getStddev(localRange, localAvg);
        this.stddev2 = 2.0 * Math.sqrt(((dev*dev) * len + (val*val)) / (len + 1));
        if (--this.stddev2IgnoreLocalDev < 0) {
            this.stddev2 = (9.0 * this.stddev2 + this.stddev2Local) / 10.0;
        }
        this.stddevSamples += 1;
    },
    adjustPanSlider: function() {
        var minValue = 0;

        visSteps = Math.max(
                Math.min(this.stepRange, DSS.addon.Metering.chartMajorTicks[this.resolution])
                - this.zoomLevel, 3);
        minValue = Math.ceil(visSteps / this.stepRange * 100);
        if (!minValue) {
            this.panSlider.setDisabled(true);
        } else {
            this.panSlider.setDisabled(false);
            this.panSlider.setMinValue(minValue);
        }
    },
    resetPanSlider: function() {
        var panSlider = this.panSlider
            minValue = 0;

        this.isShifted = false;
        panSlider.suspendEvents();
        this.adjustPanSlider();
        panSlider.setValue(100, false);
        panSlider.resumeEvents();
    },
    onPanSliderChange: function(slider, newVal) {
        var store = this.getMeteringStore(),
            count = store.getCount(),
            idx = Math.min(((count / 100) * newVal) >> 0, count - 1),
            chart = this.getMeteringChart(),
            axisX = chart.axes.get(1),
            axisY = chart.axes.get(0),
            rangeChange = false;

        if (count === 0) {
            return;
        }
        this.isShifted = (newVal < slider.maxValue);
        if (this.setTimeAxisRange(axisX, store.first().get('date'), store.getAt(idx).get('date'))) {
            rangeChange = true;
        }
        this.setCurrentNumericAxisMax(axisX, axisY, store);
        // HACK: refresh the store instead of the chart to redraw the whole line
        // otherwise only the part of the line that was previously there will appear
        store.fireEvent('refresh', store);
        // chart.refresh();
        if (rangeChange) {
            this.fireEvent('chartrangechange');
        }
    },
    setCurrentNumericAxisMax: function(axisX, axisY, store) {
        var count = store.getCount(),
            data = store.data,
            startX = axisX.fromDate,
            endX = axisX.toDate,
            bestIdx = 0,
            idx = (count >> 1), // divide by 2 (round odd down)
            maxVal = 0,
            step = idx,
            val,
            i;

        // find first store index that is currently displayed
        do {
            step >>= 1;
            val = data.getAt(idx).get('date');
            if (val < startX) {
                idx += step;
            } else if (val > startX) {
                bestIdx = idx;
                idx -= step;
            } else {
                bestIdx = idx;
                break;
            }
        } while (step > 1);
        // Hackier than above but much faster: unfortunately doesn't work without refresh after setTimeAxisRange()
        // and refresh takes longer than the iteration above
        // bestIdx = this.getMeteringChart().series.getAt(0).items[0].storeItem.index;

        // find max value between first and last displayed store indices
        i = bestIdx;
        while (true) {
            val = data.getAt(i++);
            if (!val || val.get('date') > endX) {
                break;
            }
            val = val.get('value');
            if (val > maxVal) {
                maxVal = val;
            }
        }
        this.setNumericAxisMax(axisY, maxVal);
    },
    /**
     * calculate the range to display on the y axis: ensure that this value
     * is divisible by 10 (for 2 digit numbers), 100 (for 3 digit), etc
     * Will need a chart refresh to be rendered
     */
    setNumericAxisMax: function(axis, value) {
        var max = Math.max(DSS.addon.Metering.YAxisDefaultMax, value * 1.2),
            log = Math.ceil(Math.log(max + 1) / Math.LN10); // Math.log10(max)

        max += log - (max % log); // add remainder to next round number
        axis.maximum = max;
    },
    getTimeAxisStepRange: function(axis, first, last, outValues) {
        var factor,
            toDate,
            tmp;

        outValues = outValues || {};
        if (!axis.step) {
            factor = 1000 * 60 * 60; // One hour
        } else {
            switch (axis.step[0]) {
                case Ext.Date.MINUTE:
                    factor = Math.ceil(1000 * 60 * axis.step[1]);
                    break;
                case Ext.Date.HOUR:
                    factor = Math.ceil(1000 * 60 * 60 * axis.step[1]);
                    break;
                case Ext.Date.DAY:
                    factor = Math.ceil(1000 * 60 * 60 * 24 * axis.step[1]);
                    break;
                case Ext.Date.WEEK:
                    factor = Math.ceil(1000 * 60 * 60 * 24 * 7 * axis.step[1]);
                    break;
                // MONTH and YEAR are special cases
                case Ext.Date.MONTH:
                    // assume 30-day months, that should be precise enough to calc the step-range
                    factor = Math.ceil(1000 * 60 * 60 * 24 * 30 * axis.step[1]);
                    // set toDate to first of step-th added month
                    tmp = axis.step[1] - (last.getMonth() % axis.step[1]);
                    toDate = Ext.Date.add(last, Ext.Date.MONTH, tmp);
                    toDate.setDate(1);
                    toDate.setHours(0);
                    toDate.setMinutes(0);
                    toDate.setSeconds(0);
                    toDate.setMilliseconds(0);
                    outValues.toDate = toDate;
                    outValues.factor = factor;
                    return Math.ceil((toDate.getTime() - first.getTime()) / factor);
                case Ext.Date.YEAR:
                    // assume 365-day years, that should be precise enough to calc the step-range
                    factor = Math.ceil(1000 * 60 * 60 * 24 * 365 * axis.step[1]);
                    // set toDate to first of step-th added year
                    tmp = axis.step[1] - (last.getFullYear() % axis.step[1]);
                    toDate = Ext.Date.add(last, Ext.Date.YEAR, tmp);
                    toDate.setMonth(0);
                    toDate.setDate(1);
                    toDate.setHours(0);
                    toDate.setMinutes(0);
                    toDate.setSeconds(0);
                    toDate.setMilliseconds(0);
                    outValues.toDate = toDate;
                    outValues.factor = factor;
                    return Math.ceil((toDate.getTime() - first.getTime()) / factor);
            }
        }

        // round toDate to next major tick
        toDate = Ext.Date.add(last, Ext.Date.MILLI, factor - (last.getTime() % factor));
        outValues.toDate = toDate;
        outValues.factor = factor;
        return Math.ceil((toDate.getTime() - first.getTime()) / factor);
    },
    /**
     * Set the boundaries on the time axis.
     * These are determined by the first and last data points and the current
     * zoom level. Zooming in always occurs towards the last data point.
     * While zooming out extends towards the first data point.
     * The chart must be refresh manually for the results to be rendered
     * @param axis {Object} The time axis to modify
     * @param first {Object} The first data point (Date object)
     * @param last {Object} The last data point (Date object)
     * @return {Boolean} True if the range changed, false if it remains
     */
    setTimeAxisRange: function(axis, first, last) {
        var oldToDate = axis.toDate && axis.toDate.getTime(),
            oldFromDate = axis.fromDate && axis.fromDate.getTime(),
            outValues = { toDate: null, factor: 0.0 },
            toDate,
            factor,
            stepRange;

        if (first === undefined) {
            // could use dSS time from /cgi-bin/dateinfo.sh
            // but doesn't really matter since the graph will adjust correctly
            // as soon as data is provided
            first = new Date();
        }
        if (last === undefined) {
            last = new Date();
        }
        stepRange = this.getTimeAxisStepRange(axis, first, last, outValues);
        this.stepRange = stepRange;
        toDate = outValues.toDate;
        factor = outValues.factor;
        // min 3 steps, max (chartMajorSteps - zoomLevel) on time axis
        stepRange = Math.max(
                Math.min(stepRange, DSS.addon.Metering.chartMajorTicks[this.resolution])
                - this.zoomLevel, 3);

        if (axis.step) {
            switch (axis.step[0]) {
                case Ext.Date.MONTH:
                    axis.fromDate = Ext.Date.add(toDate, Ext.Date.MONTH, -(stepRange));
                    axis.toDate = toDate;
                    return !(toDate.getTime() === oldToDate && axis.fromDate.getTime() === oldFromDate);
                case Ext.Date.YEAR:
                    axis.fromDate = Ext.Date.add(toDate, Ext.Date.YEAR, -(stepRange));
                    axis.toDate = toDate;
                    return !(toDate.getTime() === oldToDate && axis.fromDate.getTime() === oldFromDate);

            }
        }
        axis.fromDate = Ext.Date.add(toDate, Ext.Date.MILLI, -(stepRange * factor));
        axis.toDate = toDate;
        return !(toDate.getTime() === oldToDate && axis.fromDate.getTime() === oldFromDate);
    },

    /**
     * {Array} Array of DiffSprite. Stores jump markers for power spikes.
     */
    jumpMarkers: [],

    /**
     * {Array} Array of tuples [AjaxRequest id, jumpMarker]
     */
    lastCalledScenesRequests: [],

    /**
     * Add a new marker to the given store item.
     * The text is displayed in the chart
     */
    addJumpMarker: function(storeItem, diff, sceneEvent) {
        var chart = this.getMeteringChart(),
            eventName = sceneEvent && sceneEvent.eventName,
            bubble = Ext.create('DSS.addon.Metering.DiffSprite', {
                chart: chart,
                diff: diff,
                unit: this.unit,
                sceneEvent: sceneEvent,
                storeItem: storeItem,
                position: this.getPositionForStoreItem(storeItem)
            });

        this.jumpMarkers.push(bubble);
        return bubble;
    },
    onChartSelection: function(chart, sel) {
        var me = this,
            store = me.getMeteringStore(),
            idxStart,
            idxEnd,
            items,
            avg = 0.0,
            tot = 0.0,
            totUnit = 'Wh',
            sum = 0.0;

        if (sel.width === 0) {
            return;
        }

        idxStart = store.indexOf(me.getStoreItemForPosition(sel.x));
        if (idxStart === -1) {
            idxStart = 0;
        }
        idxEnd = store.indexOf(me.getStoreItemForPosition(sel.x + sel.width));
        if (idxEnd === -1) {
            idxEnd = store.getCount() -1;
        }
        items = store.getRange(idxStart, idxEnd);
        if (items.length > 0) {
            Ext.Array.each(items, function(i) { sum += i.get('value'); });
            avg = Math.round(sum / items.length);
            if (me.chartType === 'consumption') {
                tot = avg * (items[items.length -1].get('date') - items[0].get('date')) / (1000 * 60 * 60);
            } else {
                tot = sum;
            }
            if (tot > 1000) {
                tot = Math.round(tot / 100) / 10;
                totUnit = 'kWh';
            } else {
                tot = Math.round(tot);
            }
        }
        me.statsBox.setStats(tot.toString(), totUnit, avg.toString(), me.unit);
    },
    hideChartSelectionMask: function() {
        this.statsBox.clearStats();
        this.getMeteringChart().mask.hide();
    },
    getStoreItemForPosition: function(x) {
        if (this.getMeteringChart().series.first().items.length < 2) {
            return undefined;
        }
        var chart = this.getMeteringChart(),
            items = chart.series.first().items, // Careful items is private
            lastPtX = parseFloat(items[items.length -1].point[0]),
            firstPtX = parseFloat(items[0].point[0]),
            width = lastPtX - firstPtX,
            bbox = chart.getBox(),
            xCoord = x - bbox.x + chart.chartBBox.x, // Careful chartBBox is private
            c = (xCoord - firstPtX) / width,
            i = c < 1.0
                ? (c < 0.0 ? 0 : (c * items.length) >> 0)
                : items.length -1,
            spacing = (width / items.length) * 0.55,
            p = parseFloat(items[i].point[0]),
            dir = c <= 0.0 || c >= 1.0 || Math.abs(xCoord - p) <= spacing
                    ? 0
                    : p + spacing < xCoord ? 1 : -1,
            start = dir > 0 ? i : 0,
            end = dir > 0 ? items.length -1 : i,
            attr;

        if (dir === 0) {
            return items[i].storeItem;
        }

        // This wouldn't normally be required but since the items aren't
        // perfectly equispaced (resolution is 1s but query only every 2s)
        // it usually takes a 0-6 iterations
        do {
            i = (start + end) >> 1;
            p = parseFloat(items[i].point[0]);
            if (p - spacing > xCoord) {
                end = i;
            } else if (p + spacing < xCoord) {
                start = i;
            } else {
                break;
            }
        } while (Math.abs(xCoord - p) > spacing && end - start > 1);
        if (items[i].sprite) {
            // We can check the correct boundaries with column graphs to fix off-by-ones
            // Careful, sprite is private
            attr = items[i].sprite.attr;
            if (xCoord < attr.x && i > 0) {
                return items[i-1].storeItem;
            } else if (xCoord > attr.x + attr.width && i < items.length -1) {
                return items[i+1].storeItem;
            }
        }
        return items[i].storeItem;
    },
    getPositionForStoreItem: function(storeItem) {
        var items = this.getMeteringChart().series.first().items, // Careful items is private
            i;

        for (i = items.length -1; i >= 0; --i) {
            if (storeItem === items[i].storeItem) { // Careful again: private storeItem
                return items[i].point;
            }
        }
        return null;
    },
    /**
     * Update all markers with new item positions.
     * Necessary if a new item is added to the chart
     */
    updateMarkers: function() {
        var me = this;
        Ext.Array.each(this.jumpMarkers, function(m) {
            m.updatePosition(me.getPositionForStoreItem(m.storeItem));
        });
    },
    clearMarkers: function() {
        Ext.Array.each(this.jumpMarkers, function(m) { m.destroy(); });
        this.jumpMarkers = [];
        this.lastCalledScenesRequests = [];
    }
});

Ext.define('DSS.addon.Metering.DiffSprite', {
    extend: 'Ext.draw.CompositeSprite',

    constructor: function(config) {
        var chart = config.chart,
            hasSceneEvent = !!config.sceneEvent,
            evt,
            addSpriteToSurface = function(s) { chart.surface.add(s); };

        config = Ext.applyIf(config, {
            sprites: {},
            surface: chart.surface,
            timebomb: 10 * 1000,
            type: 'path',
            listeners: {
                click: function() { this.close(); }
            }
        });
        this.callParent(arguments);
        this.addEvents('sceneeventchange',
                       'storeitemchange');
        evt = hasSceneEvent ? this.truncateEventStrings(config.sceneEvent) : null;

        this.sprites.frameSprite = Ext.create('Ext.draw.Sprite', {
            type: 'path',
            path: "M 0 10 L 0 40 Q 0 50 10 50 L 170 50 L 200 60 L 180 40 L 180 10 Q 180 0 170 0 L 10 0 Q 0 0 0 10 Z",
            stroke: '#00523b',
            'stroke-width': 3,
            fill: '#baefd1',
            opacity: 0.9,
            x: 0,
            y: 0,
            zIndex: 7000
        });

        this.sprites.diffSprite = Ext.create('Ext.draw.Sprite', {
            type: 'text',
            text: this.getDiffString() + (hasSceneEvent && evt.eventName
                                    ? ' - ' + evt.eventName
                                    : ''),
            x: 5,
            y: 35,
            zIndex: 7000
        });

        this.sprites.sceneSprite = Ext.create('Ext.draw.Sprite', {
            type: 'text',
            text: hasSceneEvent && evt.sceneName || '',
            x: 5,
            y: 15,
            zIndex: 7000
        });

        chart.surface.add(this);
        this.add(this.sprites.frameSprite);
        this.add(this.sprites.diffSprite);
        this.add(this.sprites.sceneSprite);
        // need to add all sprites to the surface
        this.each(addSpriteToSurface);
        if (hasSceneEvent) {
            this.createIconSprites(config.sceneEvent);
            if (evt.truncated) {
                this.createTooltip();
            }
        } else if (config.timebomb > 0) {
            this.timebomb = setTimeout(Ext.bind(this.close, this), config.timebomb);
        }

        if (this.position) {
            this.updatePosition(this.position);
        }
    },
    getDiffString: function() {
        return (this.diff >= 0 ? ('+' + this.diff.toFixed(0))
                                : this.diff.toFixed(0))
                + this.unit;
    },
    getDiff: function() {
        return this.diff;
    },
    /** @private */
    createIconSprites: function(sceneEvent) {
        if (sceneEvent.originIcon) {
            this.sprites.originIconSprite = Ext.create('Ext.draw.Sprite', {
                type: 'image',
                width: 16,
                height: 16,
                x: 140,
                y: 30,
                zIndex: 7000,
                src: sceneEvent.originIcon
            });
            this.add(this.sprites.originIconSprite);
            this.chart.surface.add(this.sprites.originIconSprite);
        }
        if (sceneEvent.sceneIcon) {
            this.sprites.sceneIconSprite = Ext.create('Ext.draw.Sprite', {
                type: 'image',
                width: 16,
                height: 16,
                x: 155,
                y: 30,
                zIndex: 7000,
                src: sceneEvent.sceneIcon
            });
            this.add(this.sprites.sceneIconSprite);
            this.chart.surface.add(this.sprites.sceneIconSprite);
        }
    },
    /** @private */
    truncateEventStrings: function(sceneEvent) {
        var sceneName,
            zoneName,
            deviceName,
            locStr,
            len,
            remLen,
            truncated = false,
            truncateString = function(s, length) {
                if (length <= 0) { return s; }
                truncated = true;
                return s.substr(0, s.length - length -1) + '…';
            },
            ret = {};

        if (sceneEvent.sceneName) {
            if (sceneEvent.zoneId === 0) {
                ret.sceneName = sceneEvent.sceneName;
            } else {
                if (sceneEvent.originDsid) {
                    locStr = _("%device in %zone %scene");
                    len = sceneEvent.sceneName.length
                            + sceneEvent.zoneName.length
                            + sceneEvent.originName.length
                            + locStr.length -7 -5 -6; // length of %foo variables
                    remLen = ((len - 29) / 2 +1) >> 0;
                    deviceName = truncateString(sceneEvent.originName, remLen);
                    zoneName = truncateString(sceneEvent.zoneName, remLen);
                    ret.sceneName = locStr.replace('%device', deviceName)
                                          .replace('%zone', zoneName)
                                          .replace('%scene', sceneEvent.sceneName);
                } else {
                    locStr = _("%scene in %zone");
                    len = sceneEvent.sceneName.length
                            + sceneEvent.zoneName.length
                            + locStr.length -6 -5; // length of %foo variables
                    zoneName = truncateString(sceneEvent.zoneName, len - 29);
                    ret.sceneName = locStr.replace('%scene', sceneEvent.sceneName)
                                          .replace('%zone', zoneName);
                }
            }
        } else if (sceneEvent.eventName) {
            locStr = _("for %device in %zone");
            len = this.getDiffString().length
                    + sceneEvent.originName.length
                    + sceneEvent.zoneName.length
                    + locStr.length -7 -5 // length of %foo variables
                    + 3; // ' - ' separator
            ret.eventName = truncateString(sceneEvent.eventName, sceneEvent.eventName.length - 29);
            zoneName = truncateString(sceneEvent.zoneName, len - 29);
            ret.sceneName = locStr.replace('%device', sceneEvent.originName)
                                  .replace('%zone', zoneName);
        }
        ret.truncated = truncated;
        return ret;
    },
    setSceneEvent: function(sceneEvent) {
        var str = this.truncateEventStrings(sceneEvent);
        this.sceneEvent = sceneEvent;
        clearTimeout(this.timebomb);
        if (str.eventName) {
            this.sprites.diffSprite.setAttributes(
                    { text: this.getDiffString() + ' - ' + str.eventName }, true);
        }
        this.sprites.sceneSprite.setAttributes({ text: str.sceneName }, true);
        this.createIconSprites(sceneEvent);
        if (str.truncated) {
            this.createTooltip();
        }
        this.fireEvent('sceneeventchange', this, sceneEvent);
    },
    updatePosition: function(newPoint) {
        if (newPoint) {
            this.setAttributes({ translate: {
                x: newPoint[0] - 204,
                y: newPoint[1] - 62
            }}, false);
            this.show(true);
        } else {
            this.hide(true);
        }
    },
    updateStoreItem: function(newPoint, storeItem, diff) {
        this.storeItem = storeItem;
        this.diff = diff;
        this.sprites.diffSprite.setAttributes({ text: this.getDiffString() }, !newPoint);
        if (newPoint) {
            this.show(true);
            this.animate({ to: { translate: {
                x: newPoint[0] - 204,
                y: newPoint[1] - 62
            }}});
        } else {
            this.hide(true);
        }
        this.fireEvent('storeitemchange', this, storeItem);
    },
    createTooltip: function() {
        var me = this,
            evt = this.sceneEvent,
            sceneName;

        if (evt.sceneName) {
            if (evt.zoneId === 0) {
                sceneName = evt.sceneName;
            } else {
                if (evt.originDsid) {
                    sceneName = _("%device in %zone %scene")
                                    .replace('%device', evt.originName)
                                    .replace('%zone', evt.zoneName)
                                    .replace('%scene', evt.sceneName);
                } else {
                    sceneName = _("%scene in %zone")
                                    .replace('%scene', evt.sceneName)
                                    .replace('%zone', evt.zoneName);
                }
            }
        } else if (evt.eventName) {
            sceneName = _("for %device in %zone")
                            .replace('%device', evt.originName)
                            .replace('%zone', evt.zoneName);
        }

        this.tt = Ext.create('Ext.tip.ToolTip', {
            target: this.sprites.frameSprite.el,
            html: sceneName + '<br>' + this.getDiffString()
                    + (evt.eventName ? ' - ' + evt.eventName : ''),
            renderTo: Ext.getBody()
        });
        this.sprites.frameSprite.on('destroy', function() { Ext.destroy(me.tt); });
    },
    close: function() {
        var me = this,
            frame = this.sprites.frameSprite;
        // careful attr is private, dirtyHidden protected. Alternative might be to check for x===0
        if ((frame.attr.hidden === undefined || frame.attr.hidden) && !frame.dirtyHidden) {
            this.destroy();
            return;
        }
        this.animate({
            to: { opacity: 0 },
            callback: {
                fn: me.destroy,
                scope: me
            }
        });
    }
});

/*######################################
 MeteringTheme
 ######################################*/

Ext.define('Ext.chart.theme.MeteringTheme', {
    extend: 'Ext.chart.theme.Base',

    constructor: function(config) {
        var baseColor = '#005138',
            fill = '#00523b';

        this.callParent([Ext.apply({
            axis: {
                fill: baseColor,
                stroke: baseColor
            },
            axisLabelLeft: {
                fill: baseColor
            },
            axisLabelBottom: {
                fill: baseColor
            },
            axisTitleLeft: {
                fill: baseColor
            },
            axisTitleBottom: {
                fill: baseColor
            },
            colors: [fill],
            markerThemes: [{
                fill: baseColor,
                type: 'circle'
            }]
        }, config)]);
    }
});
